#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <unistd.h>

#include "crypt.h"
#include "dnet.h"

typedef enum { false = 0, true = 1 } bool;

bool key_number(uint32_t g, uint8_t n, uint64_t p, uint64_t *R)
{
	/*
	 * Cette fonction doit calculer R comme décrit dans l'algo Diffie Hellman
	 * Elle doit retourner true si le calcul n'a pas fait d'overflow, faux sinon
	 *
	 */
	*R = 0;
	return true;
}

uint32_t key_gen(uint64_t R, uint8_t n, uint64_t p)
{
	/*
	 * Cette fonction doit calculer K comme décrit dans l'algo Diffie Hellman
	 * Elle doit retourner 0 si le calcul fait un overflow, la clé sinon
	 */
	return 0;
}

struct DH_client_msg
{
	uint16_t g;
	uint64_t p;
	uint64_t A;
};

struct DH_server_msg
{
	uint64_t B;
};

uint32_t DH_client(int sock)
{
	struct DH_client_msg cmsg;
	struct DH_server_msg smsg;
	uint32_t ret;
	uint8_t a;

	/* generate the first message */
	do {
		cmsg.g = rand() % 8;
		a = rand() % 8;
		cmsg.p = (rand() * rand()) % UINT_MAX;
	} while (cmsg.g <= 1 || a <= 1 || cmsg.p <= 1 ||
		 key_number(cmsg.g, a, cmsg.p, &cmsg.A) == false);

	if (write(sock, &cmsg, sizeof(cmsg)) < sizeof(cmsg))
		perror("write failed");
	
	printf("DH_client: write done\n");

	if (read(sock, &smsg, sizeof(smsg)) < sizeof(smsg))
		perror("read failed");
	
	printf("DH_client: read done\n");

	ret = key_gen(smsg.B, a, cmsg.p);
	if (ret == 0)
		return DH_client(sock);
	else
		return ret;
}

uint32_t DH_server(int sock)
{
	struct DH_client_msg cmsg;
	struct DH_server_msg smsg;
	uint32_t ret;
	uint8_t b;

	if (read(sock, &cmsg, sizeof(cmsg)) < sizeof(cmsg))
		perror("read failed");
	printf("DH_server: read done\n");

	do {
		b = rand() % 8;
	} while (b <= 1 || !key_number(cmsg.g, b, cmsg.p, &smsg.B));

	if (write(sock, &smsg, sizeof(smsg)) < sizeof(smsg))
		perror("write failed");
	
	printf("DH_server: write done\n");

	ret = key_gen(cmsg.A, b, cmsg.p);
	if (ret == 0)
		return DH_client(sock);
	else
		return ret;
}

void printhex(const char *msg, const char *src, size_t src_len)
{
	int i;
	printf("%s", msg);
	for (i = 0; i < src_len; i++)
		printf("%.2x ", (unsigned char) src[i]);
	printf("\n");
}

int wait_for_client(uint16_t port)
{
	int sock_bind;
	char *c_ip = NULL;
	int c_port;

	sock_bind = open_tcp_server(port, 1, NULL);
	return wait_tcp_connection(sock_bind, &c_ip, &c_port);
}

int main(int argc, char **argv)
{
	const char msg[] = "message à chiffrer";
	char cryptogram[100] = { 0 };
	char msg_decrypted[100];
	uint32_t K;
	uint16_t port;
	int sock = -1;
	int ret;

	srand(time(NULL));

	if (argc != 3)
		exit(1);

	port = atoi(argv[2]);

	if (strcmp(argv[1], "server") == 0)
	{
		sock = wait_for_client(port);
		if (sock <= 0) {
			fprintf(stderr, "Listening error\n");
			return 2;
		}
		K = DH_server(sock);

		printhex("Key = ", (const char *)&K, sizeof(K));

		read(sock, cryptogram, strlen(msg) + 1);
		printhex("Le cryptogramme est      : ", cryptogram, strlen(msg));

		decrypt(cryptogram, strlen(msg) + 1, msg_decrypted, (const char*)&K, 4);
		printhex("Le message déchiffré est : ", msg_decrypted, strlen(msg));

		ret = strncmp(msg, msg_decrypted, strlen(msg)) == 0;
		if (ret)
			printf("Positive match: Bravo :)\n");
		else
			printf("Procédure de chiffrement invalide: Try again :(\n");
	}
	else if (strcmp(argv[1], "client") == 0)
	{
		sock = open_tcp_client(port, "127.0.0.1", 2);
		if (sock <= 0) {
			fprintf(stderr, "Connection error\n");
			return 2;
		}
		K = DH_client(sock);

		printhex("Key = ", (const char *)&K, sizeof(K));

		printhex("Envoi du message    : ", msg, strlen(msg));
		crypt(msg, strlen(msg) + 1, cryptogram, (const char*)&K, 4);
		printhex("Le cryptogramme est : ", cryptogram, strlen(msg));
		write(sock, cryptogram, strlen(msg) + 1);
	}
	else
		fprintf(stderr, "Invalid mode\n");

	if (sock > 0)
		close_tcp_connection(sock);
	
	return -1;
}
